from ...linear.sparse import SparseArray


def test_init():
    sp = SparseArray(10, 40.0)
    assert sp[10] == 40.0
    assert sp[0] == 0.0
    assert sp[11] == 0.0


def test_setitem():
    sp = SparseArray(10, 40.0)
    assert sp[10] == 40.0
    assert sp[0] == 0.0
    assert sp[11] == 0.0
    sp[0] = 12.0
    assert sp[0] == 12.0
    assert sp[10] == 40.0
    sp[0] = 14.0
    assert sp[0] == 14.0
    assert sp[10] == 40.0
    sp[52] = 6.0
    assert sp[0] == 14.0
    assert sp[1] == 0.0
    assert sp[10] == 40.0
    assert sp[52] == 6.0

def test_clone():
    sp1 = SparseArray(10, 40.0)
    sp2 = SparseArray(200, 2.)
    sp1 << sp2
    assert sp1[200] == 2.0
    assert sp2[200] == 2.
    sp1[2] = 2.
    assert sp1[2] == 2.
    assert 2 not in sp2
 



from libc.string cimport memcpy
from libc.stdlib cimport calloc, free
from ..typedefs cimport len_t

from .eg cimport Example

cdef class Minibatch:
    def __cinit__(self, nr_class=None, widths=None, batch_size=0):
        if widths is None and nr_class is None:
            nr_class = 1
        if widths is None:
            widths = [nr_class]
        self.c = NULL
        if widths != None:
            c_widths = <len_t*>calloc(len(widths), sizeof(len_t))
            for i, width in enumerate(widths):
                c_widths[i] = width
            self.c = new MinibatchC(c_widths, len(widths), batch_size)
            free(c_widths)
        else:
            self.c = new MinibatchC(NULL, 0, batch_size)

    def __dealloc__(self):
        if self.c != NULL:
            del self.c
        self.c = NULL

    def __len__(self):
        return self.c.i

    def __getitem__(self, int i):
        cdef Example eg = Example(nr_class=self.nr_class, nr_feat=self.c.nr_feat(i))
        memcpy(eg.c.features,
            self.c.features(i), eg.nr_feat * sizeof(eg.c.features[0]))
        memcpy(eg.c.scores,
            self.c.scores(i), eg.c.nr_class * sizeof(eg.c.scores[0]))
        memcpy(eg.c.costs,
            self.c.costs(i), eg.c.nr_class * sizeof(eg.c.costs[0]))
        memcpy(eg.c.is_valid,
            self.c.is_valid(i), eg.c.nr_class * sizeof(eg.c.is_valid[0]))
        return eg

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    @property
    def widths(self):
        return [self.c.widths[i] for i in range(self.c.nr_layer)]

    @property
    def nr_class(self):
        return self.c.nr_out()

    def guess(self, i):
        return self.c.guess(i)

    def best(self, i):
        return self.c.best(i)

    def loss(self, i):
        return 1.0 - self.c.scores(i)[self.c.best(i)]




# cython: infer_types=True
cimport cython

from six import integer_types


cdef ExampleC init_eg(Pool mem, int nr_class=0, int nr_atom=0, int nr_feat=0, widths=None):
    if widths is None:
        widths = [nr_class]
    if nr_class == 0:
        nr_class = widths[-1]

    cdef ExampleC eg
    eg.nr_class = nr_class
    eg.nr_atom = nr_atom
    eg.nr_feat = nr_feat

    eg.scores = <weight_t*>mem.alloc(nr_class, sizeof(eg.scores[0]))
    eg.costs = <weight_t*>mem.alloc(nr_class, sizeof(eg.costs[0]))
    eg.atoms = <atom_t*>mem.alloc(nr_atom, sizeof(eg.atoms[0]))
    eg.features = <FeatureC*>mem.alloc(nr_feat, sizeof(eg.features[0]))
        
    eg.is_valid = <int*>mem.alloc(nr_class, sizeof(eg.is_valid[0]))
    for i in range(eg.nr_class):
        eg.is_valid[i] = 1
    return eg
    

cdef class Example:
    def __init__(self, int nr_class=0, int nr_atom=0, int nr_feat=0):
        self.mem = Pool()
        self.c = init_eg(self.mem, nr_class=nr_class, nr_atom=nr_atom, nr_feat=nr_feat)

    def fill_features(self, int value, int nr_feat):
        for i in range(nr_feat):
            self.c.features[i].i = value
            self.c.features[i].key = value
            self.c.features[i].value = value

    def fill_atoms(self, atom_t value, int nr_atom):
        for i in range(self.c.nr_atom):
            self.c.atoms[i] = value

    def fill_scores(self, weight_t value, int nr_class):
        for i in range(self.c.nr_class):
            self.c.scores[i] = value

    def fill_is_valid(self, int value, int nr_class):
        for i in range(self.c.nr_class):
            self.c.is_valid[i] = value
   
    def fill_costs(self, weight_t value, int nr_class):
        for i in range(self.c.nr_class):
            self.c.costs[i] = value

    def reset(self):
        self.fill_features(0, self.c.nr_feat)
        self.fill_atoms(0, self.c.nr_atom)
        self.fill_scores(0, self.c.nr_class)
        self.fill_costs(0, self.c.nr_class)
        self.fill_is_valid(1, self.c.nr_class)
   
    property features:
        def __get__(self):
            for i in range(self.nr_feat):
                yield self.c.features[i]
        def __set__(self, features):
            cdef weight_t value
            cdef int slot
            if isinstance(features, dict):
                feats_dict = features
                features = []
                for key, value in feats_dict.items():
                    if isinstance(key, integer_types):
                        slot = 0
                    else:
                        slot, key = key
                    features.append((slot, key, value))
            self.nr_feat = len(features)
            cdef feat_t feat
            for i, (slot, feat, value) in enumerate(features):
                self.c.features[i] = FeatureC(i=slot, key=feat, value=value)

    property scores:
        def __get__(self):
            return [self.c.scores[i] for i in range(self.c.nr_class)]
        def __set__(self, scores):
            if len(scores) < self.nr_class:
                self.fill_scores(0, self.nr_class)
            else:
                self.nr_class = len(scores)
            for i, score in enumerate(scores):
                self.c.scores[i] = score

    property is_valid:
        def __get__(self):
            return [self.c.is_valid[i] for i in range(self.c.nr_class)]
        def __set__(self, validities):
            assert len(validities) == self.c.nr_class
            for i, is_valid in enumerate(validities):
                self.c.is_valid[i] = is_valid

    property costs:
        def __get__(self):
            return [self.c.costs[i] for i in range(self.c.nr_class)]
        def __set__(self, costs):
            cdef weight_t cost
            assert len(costs) == self.c.nr_class
            for i, cost in enumerate(costs):
                self.c.costs[i] = cost

    property guess:
        def __get__(self):
            return VecVec.arg_max_if_true(
                self.c.scores, self.c.is_valid, self.c.nr_class)

    property best:
        def __get__(self):
            cdef int i
            cdef int best = -1
            # Account for negative costs, so don't use arg_max_if_zero
            for i in range(self.c.nr_class):
                if self.c.is_valid[i] \
                and self.c.costs[i] <= 0 \
                and (best == -1 or self.c.scores[i] > self.c.scores[best]):
                    best = i
            return best
    
    property cost:
        def __get__(self):
            return self.c.costs[self.guess]
        def __set__(self, weight_t value):
            self.c.costs[self.guess] = value
   
    property nr_class:
        def __get__(self):
            return self.c.nr_class
 
    property nr_atom:
        def __get__(self):
            return self.c.nr_atom
        def __set__(self, int nr_atom):
            self.resize_atoms(nr_atom)

    property nr_feat:
        def __get__(self):
            return self.c.nr_feat
        def __set__(self, int nr_feat):
            self.c.nr_feat = nr_feat
            #self.resize_features(nr_feat)

    property loss:
        def __get__(self):
            return 1 - self.c.scores[self.best]



from ..api import layerize
import numpy

try:
    from cupy import  get_array_module
except ImportError:
    get_array_module = lambda arr: numpy


def Pooling(*funcs, **kwargs):
    ops = kwargs['ops'] if 'ops' in kwargs else funcs[0].ops
    F = len(funcs)
    def begin_update(X_lengths, drop=0.0):
        X, lengths = X_lengths
        X, bp_dropout = ops.dropout(X, drop)
        B, O = X.shape
        pooled = ops.allocate((F, len(lengths), O))
        bp_funcs = [None] * F
        for i, func in enumerate(funcs):
            pooled[i], bp_funcs[i] = func.begin_update((X, lengths))
        def finish_update(d_pooled, sgd=None):
            d_pooled = d_pooled.reshape((len(lengths), F, O))
            d_pooled = d_pooled.transpose((1, 0, 2))
            dX = ops.allocate(X.shape)
            for i, bp_func in enumerate(bp_funcs):
                dX += bp_func(d_pooled[i])
            return dX
        pooled = pooled.transpose((1, 0, 2))
        pooled = pooled.reshape((len(lengths), F * O))
        return pooled, finish_update
    return layerize(begin_update)


@layerize
def mean_pool(X_lengths, drop=0.):
    X, lengths = X_lengths
    xp = get_array_module(X)
    output = xp.zeros((len(lengths), X.shape[1]), dtype='float32')
    start = 0
    for i, length in enumerate(lengths):
        end = start + length
        output[i] = X[start : end].mean(axis=0)
        start = end
    def finish_update(d_output, sgd=None):
        d_X = xp.zeros((X.shape[0], X.shape[1]), dtype='float32')
        start = 0
        for i, length in enumerate(lengths):
            end = start + length
            d_X[start : end] += d_output[i] / (end-start)
            start = end
        return d_X
    return output, finish_update


@layerize
def max_pool(X_lengths, drop=0.):
    X, lengths = X_lengths
    xp = get_array_module(X)
    maxes = xp.zeros((len(lengths), X.shape[1]), dtype='float32')
    start = 0
    for i, length in enumerate(lengths):
        end = start + length
        maxes[i] = X[start : end].max(axis=0)
        start = end
    def finish_update(d_maxes, sgd=None):
        d_X = xp.zeros((X.shape[0], X.shape[1]), dtype='float32')
        start = 0
        for i, length in enumerate(lengths):
            end = start + length
            d_X[start : end] += d_maxes[i] * (d_maxes[i] == maxes[i])
            start = end
        return d_X
    return maxes, finish_update
